/**
  *  Yung-Pin Cheng 10/27/98
  *    This version is implemented because original dotparser is 
  *  too dinosaur and does not scale well.
  *  This version will use less memory as posssible.
  *  To achieve this goal, we choose not to use JavaCC parser, because
  *  a huge DOT file will easily blow out the stack and cause 
  *  out of memory.
  */

import java.util.* ;
import java.io.* ;
import java.lang.* ;

class dotreader {
  
  StreamTokenizer in ;
 
  char[] queue = new char[50] ;
  String[] queuestr = new String[50] ;
  int qno = 0;
  int nodeno = 0 ;
  int edgeno = 0 ;
  Hashtable nodet ;
  Hashtable edget = new Hashtable(100);
  Hashtable redget = new Hashtable(100);
  node initnode = null ;
  String initstr = null ;
  String graphname = null ;
  
  
  char[][] syntax = {
    {'I','[','P','=','I',',','S','=','I',',','C','=','I',']',';','0' },
    {'I','[','P','=','B',',','S','=','I',',','C','=','I',']',';','0' },
    {'I','-','>','I','[','L','=','I',',','C','=','I',']',';','0' },
    {'I','-','>','I','[','L','=','I',']',';','0' },
    {'D','I','{','0'},
    {'}',';','0'}} ;
    
  final int ruleno = 6 ;
  final int TAU = 2 ;
  final int EDGE = 3 ;
  final int DIGRAPH = 4 ;
  final int NODE = 0 ;
  final int INITIALNODE = 1 ;
  final int BEGIN = 4 ;

 // constructor
public dotreader(String filename) {
  try {
    in = new StreamTokenizer(new FileReader(filename));
    in.wordChars(48,57);  // '0'-'9'
    in.wordChars(65,90);  // 'A'-'Z'
    in.wordChars(95,95);  // '_'
    in.wordChars(97,122); // 'a'-'z'
    in.wordChars(46,46);  // '.' 
  } catch (FileNotFoundException e) {
    System.out.println(e); 
  }
  nodet = new Hashtable(1000);
  edget.put("", new Integer(edgeno));
  redget.put(new Integer(0), "");
  edgeno ++ ;
}
public dotreader(String filename, int node_no) {
  try {
    in = new StreamTokenizer(new FileReader(filename));
    in.wordChars(48,57);  // '0'-'9'
    in.wordChars(65,90);  // 'A'-'Z'
    in.wordChars(95,95);  // '_'
    in.wordChars(97,122); // 'a'-'z'
    in.wordChars(46,46);  // '.' 
  } catch (FileNotFoundException e) {
    System.out.println(e); 
  }
  nodet = new Hashtable(node_no);
  edget.put("", new Integer(edgeno));
  redget.put(new Integer(0), "");
  edgeno++ ;
}

private void clearqueue() { qno = 0 ; }
private void pushqueue(char c, String sval) { 
  queue[qno] = c; 
  queuestr[qno] = sval ;
  qno ++ ;
}
private int compareSyntax() {
  int i,j, k;
  boolean matched ;
  for (i=0;i< ruleno; i++) {
    k = 0 ;
    matched = true ;
    for (j=0;j < syntax[i].length ; j++) {
      if (k < qno) {
	if (queue[k] != syntax[i][j] ) {
	  matched = false ;
	  break ;
	} else 
	  k++ ;
      } else {
	matched = false ;
	break ;
      }
    }
    if (matched) return i ;
  }
  return -1 ;
}

private char getLabel(int token) {
  switch(token) {
  case in.TT_WORD : 
    if (in.sval.compareTo("label") ==0 ) return 'L' ;
    if (in.sval.compareTo("color") == 0) return 'C' ;
    if (in.sval.compareTo("shape") == 0) return 'P' ;
    if (in.sval.compareTo("style") ==0 ) return 'S' ;
    if (in.sval.compareTo("digraph") == 0) return 'D' ;
    if (in.sval.compareTo("->") == 0 ) return '>' ;
    if (in.sval.compareTo("box") == 0) return 'B' ;
    return 'I' ;
  case '[': return '[' ;
  case ']': return ']' ;
  case '=': return '=' ;
  case ',': return ',' ;
  case ';': return ';' ;
  case '{': return '{' ;
  case '}': return '}' ;
  case '"': return 'I' ;
  case '-': return '-' ;
  case '>': return '>' ;
  } // end of switch    
  return 'X' ;
}
   
public void load() {
  firstpass();
}

private void processrules(int rule) {

  node nd ;

  switch(rule) {
  case NODE: nodet.put(queuestr[0], new node(queuestr[0],false)); break ;
  case INITIALNODE: 
    initnode = new node(queuestr[0],true);
    initstr = queuestr[0] ;
    nodet.put(queuestr[0], initnode );
    break ;
  case EDGE: 
    if ((nd = (node) nodet.get(queuestr[0])) == null) {
      System.out.println("ERROR: node "+ queuestr[0] + 
			 " has not been defined yet ");
    } else {
      // queuestr[7] is the string of label
      // queuestr[3] is the string of destination node
      if (edget.containsKey(queuestr[7]) == false) {
	edget.put(queuestr[7], new Integer(edgeno));
	redget.put(new Integer(edgeno), queuestr[7]);
	nd.addEdge(edgeno, queuestr[3]);
	edgeno++ ;
      } else
	nd.addEdge(((Integer)edget.get(queuestr[7])).intValue(), queuestr[3]);
    }
    break ;
  case TAU: 
    if ((nd = (node) nodet.get(queuestr[0])) == null) {
      System.out.println("ERROR: node "+ queuestr[0] + 
			 " has not been defined yet ");
    } else 
      nd.addEdge(0, queuestr[3]);
    break ;
  case BEGIN:
    graphname = queuestr[1] ; break ;
  }
}
private void firstpass() {
  int token ;
  boolean line ;
  int rule ;

  clearqueue();
  line = false ;
  for (;;) {
    try {
      token = in.nextToken();
      if (token == in.TT_EOF ) break ;
      pushqueue(getLabel(token), in.sval);
      if (getLabel(token) == ';') { pushqueue('0',""); line = true ; }
      if (getLabel(token) == '{') { pushqueue('0',""); line = true ; }
      if (line) {
	if ((rule= compareSyntax()) == -1 )
	  System.out.println("line "+ in.lineno() + " syntax error ");
	else {
	  if (in.lineno() % 100 == 0 ) {
	    System.out.print('\15');
	    System.out.print(" 1st pass: "+ in.lineno());
	    System.out.flush();
	  }
	  processrules(rule);
	}
	clearqueue();
	line = false ;
      }
    } catch ( IOException e ) {
      System.out.println(e);
    }
  }
}

public void toFile(String filename) {
  node nd ;
  PrintWriter out ;
  try {
    out = new PrintWriter(new FileOutputStream(filename));
    out.println(" digraph \"" + graphname + "\" {");
    Enumeration allnode = nodet.elements();
    while (allnode.hasMoreElements()) {
      nd = (node) allnode.nextElement();
      out.print(nd.toString(redget));
    }
    out.println("};");
    out.close();
  } catch ( IOException e ) {
    System.out.println(e);
  }
 
}

public String getEdge(int index) { return (String) redget.get(new Integer(index));}
public node  getNode(String name) {
  return (node) nodet.get(name);
}
      
public void clearAllFlag() {
  Enumeration e ;
  node nd ;
  e = nodet.elements();
  while (e.hasMoreElements()) {
    nd = (node) e.nextElement();
    nd.deleted = false ;
    nd.visited = false ;
  }
}
public void clearVisitFlag() {
  Enumeration e ;
  node nd ;
  e = nodet.elements();
  while (e.hasMoreElements()) {
    nd = (node) e.nextElement();
    nd.visited = false ;
  }
} 
public static void main(String args[]) {
  dotreader d = null;
  switch(args.length) {
  case 1:
    d = new dotreader(args[0]);
    d.load();
    break ;
  case 2:
    d = new dotreader(args[0],(Integer.valueOf(args[1])).intValue());
    d.load();
    break ;
  default :  System.out.println("usage: java dotreader DOT [ nodeno ] ");
    break ;
  }
  if (d != null) 
    d.toFile("ttt");
}


}
  
  

      


  
  
  


  
